import glob
import os

import piexif
from datetime import datetime, timedelta

from tqdm import tqdm

from coordinates_label_photos.coordinates.coordinates_collection import CoordinatesCollection


def list_photo_filenames(directory: str) -> 'list[str]':
    ls = glob.glob(directory)
    ls.extend(glob.glob('%s/*.jpeg' % directory))
    ls.extend(glob.glob('%s/*.jpg' % directory))
    ls.extend(glob.glob('%s/*.JPEG' % directory))
    ls.extend(glob.glob('%s/*.JPG' % directory))
    return ls


def get_photo_timestamp(
        filename: str,
        timezone_offset: str = None,
        timestamp_correction:float = None
    ) -> datetime:
    """
    ¨¨Get the timestamp from the EXIF data stored in the photo under filename.
    Add the millisecond timestamp if you have installed the GoPro Lab plugin https://gopro.github.io/labs/control/precisiontime/.
    :param filename:
    :type filename: str
    :param timezone_offset: set the photo time zone delta (format '+2:00') in case the picture is not annotated (no GPS signal)
    :type timezone_offset: str
    :param timestamp_correction: an eventual number of seconds to substracte to the recorder time stamp.
    :type timestamp_correction: float
    :rtype: datetime
    """
    exif_dict = get_photo_exif(filename)
    time = exif_dict['Exif'][piexif.ExifIFD.DateTimeOriginal].decode("utf-8")
    if piexif.ExifIFD.OffsetTimeOriginal not in exif_dict['Exif']:
        if timezone_offset is None:
            time_offset = '+0000'
        else:
            time_offset = timezone_offset
    else:
        time_offset = exif_dict['Exif'][piexif.ExifIFD.OffsetTimeOriginal].decode("utf-8")

    dt =  datetime.strptime(time + ' ' + time_offset, '%Y:%m:%d %H:%M:%S %z')
    if ('Exif' in exif_dict) and piexif.ExifIFD.SubSecTime in exif_dict['Exif']:
        sub_sec_time=float(exif_dict['Exif'][piexif.ExifIFD.SubSecTime].decode("utf-8"))
        dt = dt + timedelta(seconds=sub_sec_time/1000)

    if timestamp_correction is not None:
        dt = dt + timedelta(seconds=timestamp_correction)
    return dt


def calibrate_photo(
        photos: 'list[str]',
        track_coords: CoordinatesCollection,
        photo_timezone_offset: str = None,
        photo_timestamp_correction: float = None,
        camera_fixed_elevation: float = None,
        horizontal_accuracy: float = None,
        vertical_accuracy: float = None,
        edit_photo: bool = True
) -> CoordinatesCollection:
    photo_coords = []
    for p in tqdm(photos, desc="Calibrating photos"):
        timestamp = get_photo_timestamp(p, photo_timezone_offset, photo_timestamp_correction)
        # print('Photo %s taken at %s' % (p, timestamp))
        coords = track_coords.interpolate_position(timestamp)

        if camera_fixed_elevation is not None:
            coords.elevation = camera_fixed_elevation
        if horizontal_accuracy is not None:
            coords.horiz_accuracy = horizontal_accuracy
        if vertical_accuracy is not None:
            coords.vert_accuracy = vertical_accuracy

        if edit_photo:
            set_photo_exif(filename=p, ifd='GPS', data=coords.exif_gps())
        # print('Saved: %s\t%s' % (p, coords))
        coords.label = os.path.basename(p)
        photo_coords.append(coords)
    return CoordinatesCollection(photo_coords)


def set_photo_exif(filename: str, ifd: str, data: dict):
    exif_dict = get_photo_exif(filename)
    exif_dict[ifd] = data
    exif_bytes = piexif.dump(exif_dict)
    piexif.insert(exif_bytes, filename)


def clear_photo_exif(filename: str, ifd: str):
    """
    Removes a whole IFD branch from the exif and saves the file back
    :param filename: the image location
    :type filename: str
    :param ifd: the key of the sub IFD map
    :type ifd: str
    """
    exif_dict = get_photo_exif(filename)
    del exif_dict[ifd]
    exif_bytes = piexif.dump(exif_dict)
    piexif.insert(exif_bytes, filename)


def get_photo_exif(filename: str):
    return piexif.load(filename)



import logging
from datetime import datetime

from coordinates_label_photos.coordinates import Coordinates


class CoordinatesCollection:
    points: 'list[Coordinates]'

    def __init__(self, points: 'list[Coordinates]'):
        self.points = points

    def start_time(self):
        return self.points[0].timestamp

    def end_time(self):
        return self.points[-1].timestamp

    def add(self, other):
        self.points.extend(other.points)

    def __len__(self):
        return len(self.points)

    def interpolate_position(self, timestamp: datetime):
        if timestamp < self.start_time() or timestamp >= self.end_time():
            logging.error('timestamp %s is out of bounds [%s, %s]' % (timestamp, self.start_time(), self.end_time()))
            return None
        bracket = self.__find_bracket(timestamp)
        if (bracket[0].positioning_quality is not None) and (bracket[0].positioning_quality != 'FIX'):
            logging.warning('bracket boundary at timme %s positioning quality %s is not "FIX"' % (
            bracket[0].timestamp, bracket[0].positioning_quality))
        elif (bracket[1].positioning_quality is not None) and (bracket[1].positioning_quality != 'FIX'):
            logging.warning('bracket boundary at timme %s positioning quality %s is not "FIX"' % (
                bracket[1].timestamp, bracket[1].positioning_quality))

        alpha = (timestamp - bracket[0].timestamp).total_seconds() / (
                bracket[1].timestamp - bracket[0].timestamp).total_seconds()
        return Coordinates(
            lat=self.__interpolate(alpha, bracket[0].lat, bracket[1].lat),
            lon=self.__interpolate(alpha, bracket[0].lon, bracket[1].lon),
            elevation=self.__interpolate(alpha, bracket[0].elevation, bracket[1].elevation),
            timestamp=timestamp
        )

    def lat_lon_boundaries(self):
        """
        :return: the boundaries of the coordinates, upper left and lower right
        :rtype: a pair of Coordinates
        """
        min_lat = min([c.lat for c in self.points])
        min_lon = max([c.lon for c in self.points])
        max_lat = max([c.lat for c in self.points])
        max_lon = min([c.lon for c in self.points])
        return Coordinates(lat=max_lat, lon=max_lon), Coordinates(lat=min_lat, lon=min_lon)

    def dimensions(self):
        """
        :return: the dimensions height x width
        :rtype: a pair of float
        """
        bound = self.lat_lon_boundaries()
        dx = Coordinates(bound[0].lat, bound[0].lon).distance(Coordinates(bound[1].lat, bound[0].lon))
        dy = Coordinates(bound[0].lat, bound[0].lon).distance(Coordinates(bound[0].lat, bound[1].lon))
        return dx, dy

    def label_all(self, label: str):
        for p in self.points:
            p.label = label

    @staticmethod
    def __interpolate(alpha, v0, v1):
        return v0 + alpha * (v1 - v0)

    def __find_bracket(self, timestamp: datetime):
        i0 = 0
        i1 = len(self.points) - 1
        while i1 - i0 > 1:
            i_mid = int((i0 + i1) / 2)
            t_mid = self.points[i_mid].timestamp
            if t_mid > timestamp:
                i1 = i_mid
            else:
                i0 = i_mid
        return self.points[i0], self.points[i1],

    def __repr__(self):
        return '\n'.join([str(c) for c in self.points])



import unittest

from coordinates_label_photos.coordinates import Coordinates
from coordinates_label_photos.parsers import coords_parser
from coordinates_label_photos.photos import get_photo_timestamp


class TestLabelPhoto(unittest.TestCase):
    photo_filename = 'resources/photo-gps-timing/G0160150-tile-11.JPG'
    photo_timestamp = get_photo_timestamp(
        photo_filename,
        timezone_offset='+0300',
        timestamp_correction=-74.5
    )
    tile_coords = Coordinates(
        lon=23.13363763,
        lat=37.42803984
    )

    def test_photo_should_be_above_tile_11_pos(self):
        """
        We use a picture and correct the time with time zone + gopro offset.
        The returned position should correspond to the measured tile.
        """
        track_coords = coords_parser('resources/photo-gps-timing/ts_reach_ro.pos')

        photo_coords = track_coords.interpolate_position(self.photo_timestamp)

        got_dist = photo_coords.distance(self.tile_coords)
        self.assertLessEqual(got_dist, 1)

    #
    # def test_photo_should_be_above_tile_11_llh(self):
    #     """
    #     We use a picture and correct the time with time zone + gopro offset.
    #     The returned position should correspond to the measured tile.
    #     """
    #     track_coords = coords_parser('resources/photo-gps-timing/ts_reach_ro.LLH')
    #
    #     photo_coords = track_coords.interpolate_position(self.photo_timestamp)
    #
    #     got_dist = photo_coords.distance(self.tile_coords)
    #     self.assertLessEqual(got_dist, 1)



import argparse
import logging

import piexif
from piexif import TAGS

from coordinates_label_photos.photos import get_photo_timestamp, get_photo_exif


def main():
    logging.basicConfig(level=logging.INFO)
    parser = argparse.ArgumentParser(description='print EXIF information')

    parser.add_argument('files', nargs=argparse.REMAINDER)

    args = parser.parse_args()
    print(args.files)

    for f in args.files:
        print(f)
        print(get_photo_timestamp(f))
        exif = get_photo_exif(f)
        for p, f in [
            ['0th', piexif.ImageIFD.Model],
            ['0th', piexif.ImageIFD.DateTime],
            ['Exif', piexif.ExifIFD.SubSecTime],
            ['Exif', piexif.ExifIFD.OffsetTime],
            ['Exif', piexif.ExifIFD.DateTimeOriginal],
            ['Exif', piexif.ExifIFD.SubSecTimeOriginal],
            ['Exif', piexif.ExifIFD.OffsetTimeOriginal],
            ['Exif', piexif.ExifIFD.DateTimeDigitized],
            ['Exif', piexif.ExifIFD.SubSecTimeDigitized],
            ['Exif', piexif.ExifIFD.OffsetTimeDigitized],
        ]:
            print('%s/%s: %s' % (p, TAGS[p][f]['name'], exif[p][f].decode("utf-8")))


if __name__ == '__main__':
    main()



import csv
import logging
import os
from datetime import datetime, timezone, timedelta

import gpxpy as gpxpy

from coordinates_label_photos.coordinates import Coordinates
from coordinates_label_photos.coordinates.coordinates_collection import CoordinatesCollection


class NoParserForCoordinateFormatError:
    def __init__(self, extension: str):
        self.extension = extension

    def __str__(self):
        return 'No parser is available for %s among %s' % (self.extension, ['.llh', '.gpx'])


def coords_parser(filename) -> CoordinatesCollection:
    """
    Read coordinates from various file format
    :param filename:
    :type filename:
    :return:
    :rtype:
    """
    logging.info('Loading coordinates track from %s' % filename)
    _, file_extension = os.path.splitext(filename)
    file_extension = file_extension.lower()
    if file_extension == '.gpx':
        return gpx_parser(filename)
    else:
        if file_extension in {'.llh', '.pos'}:
            return emlid_fwf_parser(filename)
        else:
            raise NoParserForCoordinateFormatError(file_extension)


def gpx_parser(filename) -> CoordinatesCollection:
    logging.info('Loading GPX track from %s' % filename)
    with open(filename) as fd:
        gpx = gpxpy.parse(fd)

        coords = []
        for track in gpx.tracks:
            for segment in track.segments:
                for point in segment.points:
                    coords.append(Coordinates(
                        lat=point.latitude,
                        lon=point.longitude,
                        elevation=point.elevation,
                        timestamp=point.time
                    ))
        collect = CoordinatesCollection(coords)
        logging.info('Loaded GPX with %d track points between %s and %s' % (
            len(collect), collect.start_time(), collect.end_time()))
        return collect


def csv_parser(filename) -> CoordinatesCollection:
    logging.info('Loading CSV track from %s' % filename)
    with open(filename, newline='') as csvfile:
        reader = csv.DictReader(csvfile)
        coords = []

        for row in reader:
            coords.append(
                Coordinates(
                    lat=float(row['latitude_decimal_degree']),
                    lon=float(row['longitude_decimal_degree']),
                    elevation=float(row['ellipsoidal_height_m']),
                    timestamp=None,
                )
            )
        collect = CoordinatesCollection(coords)
        logging.info('Loaded CSV with %d track points between %s and %s' % (
            len(collect), collect.start_time(), collect.end_time()))
        return collect


def emlid_fwf_parser(filename) -> CoordinatesCollection:
    """
    Parse Emlid fixed width format.
    IF a .pos file, check the last comment list for GPST and eventually correct the infamous 18 seconds
    :param filename: the source file path
    :type filename: string
    :return: the list of coordinates
    :rtype: CoordinatesCollection
    """
    logging.info('Loading Emlid .LLH or .POS from %s' % filename)

    is_gpst = False
    with open(filename) as fd:
        coords = []
        for line in fd.readlines():
            if line.startswith('%'):
                if 'GPST' in line:
                    is_gpst = True
                continue
            c = emlid_fwf_line_parser(line)
            if is_gpst:
                c.timestamp = c.timestamp + timedelta(seconds=-18)
            coords.append(c)
        return CoordinatesCollection(coords)


def _parse_emlid_timestamp(text: str):
    return datetime.strptime(text + '000', '%Y/%m/%d %H:%M:%S.%f').replace(tzinfo=timezone.utc)


def emlid_fwf_line_parser(line: str) -> Coordinates:
    timestamp = _parse_emlid_timestamp(line[0:23])
    lat = float(line[25:38])
    lon = float(line[39:53])
    elev = float(line[55:64])
    sd_n = float(line[73:82])
    sd_e = float(line[82:91])
    sd_u = float(line[91:100])
    return Coordinates(
        lat=lat,
        lon=lon,
        elevation=elev,
        timestamp=timestamp,
        horiz_accuracy=max([sd_n, sd_e]) * 1.96,
        vert_accuracy=sd_u * 1.96,
        positioning_quality=__positioning_qualitiy[int(line[67:70])]
    )


__positioning_qualitiy = {
    1: 'FIX',
    2: 'FLOAT',
    3: 'SBAS',
    4: 'DGPS',
    5: 'SINGLE',
    6: 'PPP'
}



from datetime import datetime, timezone
from unittest import TestCase

from coordinates_label_photos.parsers import coords_parser


class TestParser(TestCase):
    def test_emlid_pos_parser(self):
        coords = coords_parser('resources/photo-gps-timing/ts_reach_ro.pos')

        got_timestamp = coords.start_time()
        self.assertEqual(datetime(2022, 8, 8, 6, 33, 42, tzinfo=timezone.utc), got_timestamp)

    def test_emlid_llh_parser(self):
        coords = coords_parser('resources/photo-gps-timing/ts_reach_ro.LLH')

        got_timestamp = coords.start_time()
        self.assertEqual(datetime(2022, 8, 8, 6, 34, 0, tzinfo=timezone.utc), got_timestamp)




from datetime import datetime, timezone, timedelta
from unittest import TestCase

from coordinates_label_photos.photos import get_photo_timestamp


class TestPhoto(TestCase):
    def test_get_photo_timestamp(self):
        # BTW, the picture shows a watch displaying 11:29:10, but the EXIF tag is 11:29:11
        got = get_photo_timestamp('resources/photo-gps/IMG_4967.jpeg')

        self.assertEqual(datetime(2022, 3, 6, 10, 29, 11, 880000, tzinfo=timezone.utc), got)

    def test_get_photo_timestamp_correction(self):
        # recorded EXIF data: 2022-08-08 09:36:13.145000+00:00
        # TZ +0300
        # GoPro clock running 74.3 seconds to early
        print(get_photo_timestamp(
            'resources/photo-gps-timing/G0160150-tile-11.JPG'
        ))
        got = get_photo_timestamp(
            'resources/photo-gps-timing/G0160150-tile-11.JPG',
            timezone_offset='+0300',
            timestamp_correction=-74.3
        )

        got_delta = datetime(2022, 8, 8, 6, 34, 58, 700000 + 145000, tzinfo=timezone.utc) - got
        self.assertEqual(timedelta(seconds=0), got_delta)



from datetime import datetime
from fractions import Fraction

from piexif import GPSIFD, ImageIFD
import geopy.distance


class Coordinates:
    lat: float
    lon: float
    elevation: float
    timestamp: datetime
    label: str
    horiz_accuracy: float
    vert_accuracy: float
    positioning_quality: str

    def __init__(self,
                 lat: float,
                 lon: float,
                 elevation: float = None,
                 timestamp: datetime = None,
                 label: str = None,
                 horiz_accuracy: float = None,
                 vert_accuracy: float = None,
                 positioning_quality: str = None
                 ):
        self.lat = lat
        self.lon = lon
        self.elevation = elevation
        self.timestamp = timestamp
        self.label = label
        self.horiz_accuracy = horiz_accuracy
        self.vert_accuracy = vert_accuracy
        self.positioning_quality = positioning_quality

    def distance(self, other):
        return geopy.distance.distance(
            (self.lat, self.lon),
            (other.lat, other.lon),
            ellipsoid='WGS-84'
        ).m

    def __repr__(self):
        return '(%.10f, %.10f, %s) %s %s' % (self.lat, self.lon, self.elevation, self.label, self.timestamp)

    def exif_gps(self):
        return {
            GPSIFD.GPSVersionID: (2, 0, 0, 0),
            GPSIFD.GPSAltitudeRef: 1,
            GPSIFD.GPSAltitude: _to_rational(self.elevation),
            GPSIFD.GPSDateStamp: u"2021:08:10 10:55:55",
            GPSIFD.GPSLatitudeRef: self.lat_dir(),
            GPSIFD.GPSLatitude: self.exiv_lat(),
            GPSIFD.GPSLongitudeRef: self.lon_dir(),
            GPSIFD.GPSLongitude: self.exiv_lon(),
            #GPSIFD.GPSHPositioningError: 0.4
        }

    def exiv_lat(self):
        return [_to_rational(x) for x in to_abs_deg_min_sec(self.lat)]

    def exiv_lon(self):
        return [_to_rational(x) for x in to_abs_deg_min_sec(self.lon)]

    def lat_dir(self) -> str:
        if self.lat > 0:
            return 'N'
        else:
            if self.lat < 0:
                return 'S'
            else:
                return ''

    def lon_dir(self) -> str:
        if self.lat > 0:
            return 'E'
        else:
            if self.lat < 0:
                return 'W'
            else:
                return ''


def _to_rational(number: float):
    """
    returns a numerator,denominator pair. Truncated the number to 7 decimals, to avoid  > long klimit in exif
    """
    f = Fraction(str(round(number, 7)))
    return f.numerator, f.denominator


def to_abs_deg_min_sec(value):
    """
    return the value to the absolute value in degree, minute, seconds
    """
    abs_value = abs(value)
    degrees = int(abs_value)
    rem = (abs_value - degrees) * 60
    minutes = int(rem)
    seconds = round((rem - minutes) * 60, 8)
    return degrees, minutes, seconds


def get_decimal_from_dms(dms, ref):
    """From https://developer.here.com/blog/getting-started-with-geocoding-exif-image-metadata-in-python3"""
    degrees = dms[0][0] / dms[0][1]
    minutes = dms[1][0] / dms[1][1] / 60.0
    seconds = dms[2][0] / dms[2][1] / 3600.0

    if ref in ['S', 'W']:
        degrees = -degrees
        minutes = -minutes
        seconds = -seconds

    return round(degrees + minutes + seconds, 8)


def exif_to_coordinates(exif) -> Coordinates:
    if 'GPS' not in exif:
        raise Exception('not "GPS" tag in exif image')

    exif_gps = exif['GPS']
    lat = get_decimal_from_dms(exif_gps[GPSIFD.GPSLatitude], exif_gps[GPSIFD.GPSLatitudeRef].decode('utf-8'))
    lon = get_decimal_from_dms(exif_gps[GPSIFD.GPSLongitude], exif_gps[GPSIFD.GPSLongitudeRef].decode('utf-8'))
    elev = round(exif_gps[GPSIFD.GPSAltitude][0] / exif_gps[GPSIFD.GPSAltitude][1], 2)

    timestamp = datetime.strptime(exif['0th'][ImageIFD.DateTime].decode('utf-8'), '%Y:%m:%d %H:%M:%S')

    return Coordinates(
        lat=lat,
        lon=lon,
        elevation=elev,
        timestamp=timestamp,
    )



import argparse
import logging

from coordinates_label_photos.parsers import coords_parser
from coordinates_label_photos.photos import list_photo_filenames, calibrate_photo, clear_photo_exif
from coordinates_label_photos.reports.coordinates_collection import report_image_coordinates_collections


def main():
    logging.basicConfig(level=logging.INFO)
    parser = argparse.ArgumentParser(description='Clear GPS information from photos')

    parser.add_argument(
        "--images",
        help="image directory (edited in place)",
        type=str,
        required=True,
        metavar='/path/to/images'
    )

    args = parser.parse_args()

    for p in list_photo_filenames(getattr(args, 'images')):
        clear_photo_exif(p, 'GPS')
        logging.info('Cleared GPS from %s' % p)


if __name__ == '__main__':
    main()



import argparse
import logging
import gpxpy as gpxpy
from tqdm import tqdm

from coordinates_label_photos.coordinates import exif_to_coordinates
from coordinates_label_photos.photos import list_photo_filenames, get_photo_exif


def main():
    logging.basicConfig(level=logging.INFO)
    parser = argparse.ArgumentParser(description='extract GPS coordinates from an image folder and create a gpx track')

    parser.add_argument("--images", help="image directory (edited in place)", type=str, required=True,
                        metavar='/path/to/images')
    parser.add_argument("--output", help=".gpx coordinates file", type=str, required=True, metavar='/path/to/gpx')

    args = parser.parse_args()

    photos = list_photo_filenames(getattr(args, 'images'))

    gpx = gpxpy.gpx.GPX()
    gpx_track = gpxpy.gpx.GPXTrack()
    gpx.tracks.append(gpx_track)
    gpx_segment = gpxpy.gpx.GPXTrackSegment()
    gpx_track.segments.append(gpx_segment)

    for photo in tqdm(photos, desc='Parsing images'):
        exif = get_photo_exif(photo)
        coord = exif_to_coordinates(exif)
        gpx_segment.points.append(
            gpxpy.gpx.GPXTrackPoint(coord.lat, coord.lon, elevation=coord.elevation, time=coord.timestamp))

    gpx_file = getattr(args, 'output')
    with open(gpx_file, 'w') as fd_out:
        fd_out.write(gpx.to_xml())
        fd_out.close()
        logging.info('Saved %s points in GPX %s' % (len(photos), gpx_file))


if __name__ == '__main__':
    main()



import argparse
import logging

from coordinates_label_photos.parsers import coords_parser
from coordinates_label_photos.reports.coordinates_collection import CoordinatesCollection, \
    report_image_coordinates_collections


def main():
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser(description='Plot coords tracks')

    parser.add_argument("gpx", help='label and path to .gpx, or .llh, .pos file', type=str, nargs='+',
                        metavar='label:path/to/file.gpx')
    parser.add_argument("--output", help="output image with track points comparisons", type=str, required=True,
                        metavar='/path/to/output.jpeg')

    args = parser.parse_args()

    all_coords = CoordinatesCollection([])
    for e in getattr(args, 'gpx'):
        tmp = e.split(':', maxsplit=2)
        label = tmp[0]
        gpx_file = tmp[1]
        cs = coords_parser(gpx_file)
        cs.label_all(label)
        all_coords.add(cs)
    report_image_coordinates_collections(all_coords, getattr(args, 'output'), 1000, color_by_label=True)


if __name__ == '__main__':
    main()
